//---------------------------------------------------------------------------

#ifdef _WIN32
 #include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include <GL/glew.h>
#include <GL/glut.h>

#define FB_WIDTH 640
#define FB_HEIGHT 480

//---------------------------------------------------------------------------

GLuint gWinId = 0;
GLuint gResx = 0;
GLuint gResy = 0;

typedef unsigned char byte;

#pragma pack(push)
#pragma pack(1)

typedef struct
{
    byte  identsize;          // size of ID field that follows 18 byte header (0 usually)
    byte  colourmaptype;      // type of colour map 0=none, 1=has palette
    byte  imagetype;          // type of image 0=none,1=indexed,2=rgb,3=grey,+8=rle packed

    short colourmapstart;     // first colour map entry in palette
    short colourmaplength;    // number of colours in palette
    byte  colourmapbits;      // number of bits per palette entry 15,16,24,32

    short xstart;             // image x origin
    short ystart;             // image y origin
    short width;              // image width in pixels
    short height;             // image height in pixels
    byte  bpp;               // image bits per pixel 8,16,24,32
    byte  descriptor;         // image descriptor bits (vh flip bits)

    // pixel data follows header

} TGA_HEADER;

#pragma pack(pop)

GLuint loadTGA(const char* fname)
{
	FILE* f=fopen(fname,"rb");
	if (!f) return 0;
	TGA_HEADER head;
	fread(&head,1,sizeof(TGA_HEADER),f);
	GLvoid* pixels=0;
	GLuint texID=0;
	if (head.imagetype==2) // RGB not packed
	{
		if (head.bpp==24)
		{
			pixels=(unsigned char*)malloc(head.width*head.height*3);
			fread(pixels,head.height,head.width*(head.bpp>>3),f);
			glGenTextures(1,&texID);
			glBindTexture(GL_TEXTURE_2D,texID);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);	// Linear Filtering
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);	// Linear Filtering

			if ((head.descriptor & 0x30) == 0) // dolu hlavou, BGR
				glTexImage2D(GL_TEXTURE_2D,0,GL_RGB8,head.width,head.height,0,GL_BGR,GL_UNSIGNED_BYTE,pixels);

		}
	}
	free(pixels);
	fclose(f);
	return texID;
}



GLfloat gPositions[]={
		// Front Face
		-1.0f, -1.0f,  1.0f,
		1.0f, -1.0f,  1.0f,
		1.0f,  1.0f,  1.0f,
		-1.0f,  1.0f,  1.0f,
		// Back Face
		-1.0f, -1.0f, -1.0f,
		-1.0f,  1.0f, -1.0f,
		1.0f,  1.0f, -1.0f,
		1.0f, -1.0f, -1.0f,
		// Top Face
		-1.0f,  1.0f, -1.0f,
		-1.0f,  1.0f,  1.0f,
		1.0f,  1.0f,  1.0f,
		 1.0f,  1.0f, -1.0f,
		// Bottom Face
		-1.0f, -1.0f, -1.0f,
		 1.0f, -1.0f, -1.0f,
		 1.0f, -1.0f,  1.0f,
		-1.0f, -1.0f,  1.0f,
		// Right face
		 1.0f, -1.0f, -1.0f,
		1.0f,  1.0f, -1.0f,
		 1.0f,  1.0f,  1.0f,
		 1.0f, -1.0f,  1.0f,
		// Left Face
		-1.0f, -1.0f, -1.0f,
		-1.0f, -1.0f,  1.0f,
		-1.0f,  1.0f,  1.0f,
		-1.0f,  1.0f, -1.0f
		};


GLfloat gTexCoords[]={
		// Front Face
		0.0f, 0.0f,
		1.0f, 0.0f,
		1.0f, 1.0f,
		0.0f, 1.0f,
		// Back Face
		1.0f, 0.0f,
		1.0f, 1.0f,
		0.0f, 1.0f,
		0.0f, 0.0f,
		// Top Face
		0.0f, 1.0f,
		0.0f, 0.0f,
		1.0f, 0.0f,
		1.0f, 1.0f,
		// Bottom Face
		1.0f, 1.0f,
		0.0f, 1.0f,
		0.0f, 0.0f,
		1.0f, 0.0f,
		// Right face
		1.0f, 0.0f,
		1.0f, 1.0f,
		0.0f, 1.0f,
		0.0f, 0.0f,
		// Left Face
		0.0f, 0.0f,
		1.0f, 0.0f,
		1.0f, 1.0f,
		0.0f, 1.0f};


GLuint gVertexIDs[]={
				  0,1,2,
				  2,3,0,
				  4,5,6,
				  6,7,4,
				  8,9,10,
				  10,11,8,
				  12,13,14,
				  14,15,12,
				  16,17,18,
				  18,19,16,
				  20,21,22,
				  22,23,20};

GLuint gPositionsVB;
GLuint gTexCoordsVB;
GLuint gTrianglesIB;
GLuint gTex0;

// FBO
GLuint gFB;
GLuint gFBColorTexID; // Color Texture
GLuint gFBDepthTexID; // Depth Texture
GLuint gFBdepthRB; // Depth RenderBuffer

bool loadTextures()
{
	gTex0=loadTGA("lena.tga");
	return true;
}

#define ERRORMSG(s) fprintf(stderr,"%s\n",s);fflush(stderr);


int check(GLenum status)
{
	switch (status){
		case GL_FRAMEBUFFER_COMPLETE_EXT:
			//INFOMSG("GL_FRAMEBUFFER_COMPLETE_EXT");
			return 1;
		case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
			ERRORMSG("GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
			ERRORMSG("GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT");
			break;
/*
		case GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT:
			ERRORMSG("GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT");
			break;
*/
		case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
			ERRORMSG("GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
			ERRORMSG("GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
			ERRORMSG("GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
			ERRORMSG("GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT");
			break;
		case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
			ERRORMSG("GL_FRAMEBUFFER_UNSUPPORTED_EXT");
			break;
		default:
			ERRORMSG("Unknown FBO error");
			break;
	}
	return 0;
}
   #define CHECK_FRAMEBUFFER_STATUS()                            \
      {                                                           \
        GLenum status;                                            \
        status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT); \
        check(status);                                          \
      }


bool createFBO()
{
	// Color texture for Color(draw) buffer
	glGenTextures(1,&gFBColorTexID);
	glBindTexture(GL_TEXTURE_2D,gFBColorTexID);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);	// Linear Filtering
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);	// Linear Filtering
	// initialize texture data storage
	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA8,FB_WIDTH,FB_HEIGHT,0,GL_RGBA,GL_UNSIGNED_BYTE,NULL);

	// Depth texture to get also the depth buffer into texture
	glGenTextures(1,&gFBDepthTexID);
	glBindTexture(GL_TEXTURE_2D,gFBDepthTexID);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);	// Linear Filtering
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);	// Linear Filtering
	// initialize texture data storage - use GL_DEPTH_COMPONENT16 if it does not work
	glTexImage2D(GL_TEXTURE_2D,0,GL_DEPTH_COMPONENT24,FB_WIDTH,FB_HEIGHT,0,GL_DEPTH_COMPONENT,GL_UNSIGNED_BYTE,NULL);

	// create Detph render buffer for depth-testing when using FBO
	glGenRenderbuffersEXT(1, &gFBdepthRB);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, gFBdepthRB);
	// initialize renderbuffer data storage - use GL_DEPTH_COMPONENT16 if it does not work
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, FB_WIDTH,FB_HEIGHT);

	// create FrameBuffes
	glGenFramebuffersEXT(1, &gFB);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, gFB);

	// attach Detph buffer
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_RENDERBUFFER_EXT,gFBdepthRB);
    // attach Depth texture to retrive depth buffer for future processing
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_DEPTH_ATTACHMENT_EXT,GL_TEXTURE_2D,gFBDepthTexID, 0);
    // attach Color texture to retrive color buffer for future processing
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT,GL_COLOR_ATTACHMENT0_EXT,GL_TEXTURE_2D,gFBColorTexID, 0);
	CHECK_FRAMEBUFFER_STATUS()

	// disable frame buffer
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	return true;
}

bool deleteFBO()
{
    glDeleteRenderbuffersEXT(1,&gFBdepthRB);
    glDeleteFramebuffersEXT(1,&gFB);
	return true;
}

bool createVBO()
{
// target = {GL_ARRAY_BUFFER, GL_ELEMENT_ARRAY_BUFFER, GL_PIXEL_UNPACK_BUFFER, GL_PIXEL_PACK_BUFFER}

    glGenBuffers(1, &gPositionsVB); // positions
	glBindBuffer(GL_ARRAY_BUFFER, gPositionsVB);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gPositions), gPositions,GL_STATIC_DRAW_ARB);

    glGenBuffers(1, &gTexCoordsVB); // texture coordinates
	glBindBuffer(GL_ARRAY_BUFFER, gTexCoordsVB);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gTexCoords), gTexCoords,GL_STATIC_DRAW_ARB);

	glGenBuffers(1, &gTrianglesIB); // Index Buffer
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,gTrianglesIB);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(gVertexIDs), gVertexIDs,GL_STATIC_DRAW_ARB);

	return true;
}

bool deleteVBO()
{
	glDeleteBuffers(1,&gPositionsVB);
	glDeleteBuffers(1,&gTexCoordsVB);
	glDeleteBuffers(1,&gTrianglesIB);
	return true;
}

float angleY=0.0f;
float angleX=0.0f;


void Redraw()
{
	glDisable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);
	glLoadIdentity();
	glTranslatef(0,0,-2.5);
	int time = glutGet(GLUT_ELAPSED_TIME);
	glRotatef((float)time/100.0f,0,1,1);

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, gFB);
	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
// rendering into multiple color attachements - requires shaders
//	GLuint buffers[]={GL_COLOR_ATTACHMENT0_EXT,GL_COLOR_ATTACHMENT1_EXT};
//	glDrawBuffers(2,buffers);
	glClearColor(0.3,0.1,0.4,1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glColor4f(1.0,1.0,1.0,1.0);
	glBindTexture(GL_TEXTURE_2D,gTex0);
	glEnable(GL_TEXTURE_2D);
	glBindBufferARB( GL_ARRAY_BUFFER_ARB, gPositionsVB);	// bind ARRAY BUFFER with glBuffer of vertex positions
	glVertexPointer( 3, GL_FLOAT, 0, (char *) NULL );		// Set The Vertex Pointer To bound ARRAY BUFFER
	glBindBufferARB( GL_ARRAY_BUFFER_ARB, gTexCoordsVB );	// bind ARRAY BUFFER with glBuffer of texture coordinates
	glTexCoordPointer( 2, GL_FLOAT, 0, (char *) NULL );		// Set The TexCoord Pointer To bound ARRAY BUFFER
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, gTrianglesIB);	// bind ELEMENT ARRAY BUFFER with glBuffer of indices

	glEnableClientState( GL_VERTEX_ARRAY );					// Enable Vertex Arrays
	glEnableClientState( GL_TEXTURE_COORD_ARRAY );			// Enable Texture Coord Arrays

// render indexed primitives using indices stored in bound ELEMENT ARRAY BUFFER (last parameter is NULL)
//	glDrawElements(GL_TRIANGLES,12*3,GL_UNSIGNED_INT,NULL);
	glDrawRangeElements(GL_TRIANGLES,0,23,12*3,GL_UNSIGNED_INT,NULL);

// render primitives using enabled array pointers (GL_VERTEX_ARRAY, GL_TEXTURE_COORD_ARRAY)
//	glDrawArrays(GL_QUADS,0,12*3);

	glDisableClientState( GL_VERTEX_ARRAY );				// Disable Vertex Arrays
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );			// Disable Texture Coord Arrays
// stop usinf frame buffer - return to standard rendering into window
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

	glClearColor(0.0,0.0,0.0,1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
// show rendered texture (use gFBDepthTexID to show depth buffer)
	glBindTexture(GL_TEXTURE_2D,gFBColorTexID);
	glEnable(GL_TEXTURE_2D);

	glLoadIdentity();
	glTranslatef(0,0,-1.5);
    glRotatef(angleY,0,1,0);
	glRotatef(angleX,1,0,0);

	GLfloat d=0.0;
	glColor4f(1.0,1.0,1.0,1.0);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 0.0);
		glVertex2f(-1.0, -1.0);
		glTexCoord2f(0.0, 1.0);
		glVertex2f(1.0, -1.0);
		glTexCoord2f(1.0, 1.0);
		glVertex2f(1.0, 1.0);
		glTexCoord2f(1.0, 0.0);
		glVertex2f(-1.0, 1.0);
	glEnd();

    glutSwapBuffers();
}


void ReSize(int width_in, int height_in)
{
    glViewport(0, 0, width_in, height_in);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluPerspective(90.0,1.2,0.1,5.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	gResx = width_in;
	gResy = height_in;

	Redraw();
}


void MouseMove(int x_in, int y_in)
{
	angleY=x_in;
	angleX=y_in;
}


void KeyPressed(unsigned char key_in, int x_in, int y_in)
{
    switch(key_in)
	{
		case 27:
			{
                glutDestroyWindow(gWinId);
                deleteVBO();
                deleteFBO();
			    exit(0);
			}
			break;
	}
}


int main(int argc, char **argv)
{
    //set glut window settings
    glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
	glutInitWindowSize(FB_WIDTH, FB_HEIGHT);
	glutInitWindowPosition(100, 100);
    //create glut window
    gWinId = glutCreateWindow("VBO & FBO example");

    //set glut contorling function
	glutDisplayFunc(&Redraw);
	glutIdleFunc(&Redraw);
	glutReshapeFunc(&ReSize);
	glutKeyboardFunc(&KeyPressed);
	glutMotionFunc(&MouseMove);

	if (GLEW_OK != glewInit())
    {
		fprintf(stderr,"Error glew initialization failed.\n");
    }

	loadTextures();
	createVBO();
	createFBO();

	//enable main loop
	glutMainLoop();

	return 0;
}

